package java.lang.invoke;

import java.lang.ref.SoftReference;
import java.util.Collections;
import sun.invoke.util.Wrapper;
import java.util.Arrays;
import java.util.concurrent.ConcurrentHashMap;

class LambdaFormEditor
{
    final LambdaForm lambdaForm;
    private static final int MIN_CACHE_ARRAY_SIZE = 4;
    private static final int MAX_CACHE_ARRAY_SIZE = 16;
    
    private LambdaFormEditor(LambdaForm lambdaForm) {
        this.lambdaForm = lambdaForm;
    }
    
    static LambdaFormEditor lambdaFormEditor(LambdaForm lambdaForm) {
        return new LambdaFormEditor(lambdaForm.uncustomize());
    }
    
    private LambdaForm getInCache(LambdaFormEditor.Transform key) {
        assert key.get() == null;
        Object transformCache = this.lambdaForm.transformCache;
        LambdaFormEditor.Transform transform = null;
        if ((transformCache instanceof ConcurrentHashMap)) {
            transform = (LambdaFormEditor.Transform)((ConcurrentHashMap)transformCache).get(key);
        }
        else {
            if (transformCache == null) {
                return null;
            }
            if ((transformCache instanceof LambdaFormEditor.Transform)) {
                LambdaFormEditor.Transform transform3 = (LambdaFormEditor.Transform)transformCache;
                if (transform3.equals(key)) {
                    transform = transform3;
                }
            }
            else {
                LambdaFormEditor.Transform[] array = (LambdaFormEditor.Transform[])(LambdaFormEditor.Transform[])transformCache;
                int i = 0;
                while (i < array.length) {
                    LambdaFormEditor.Transform transform2 = array[i];
                    if (transform2 == null) {
                        break;
                    }
                    if (transform2.equals(key)) {
                        transform = transform2;
                        break;
                    }
                    ++i;
                }
            }
        }
        assert !(!key.equals(transform));
        return (transform != null) ? ((LambdaForm)transform.get()) : null;
    }
    
    private LambdaForm putInCache(LambdaFormEditor.Transform withResult, LambdaForm lambdaForm) {
        withResult = withResult.withResult(lambdaForm);
        int n = 0;
        while (true) {
            Object transformCache = this.lambdaForm.transformCache;
            Label_0518: {
                if ((transformCache instanceof ConcurrentHashMap)) {
                    ConcurrentHashMap concurrentHashMap = (ConcurrentHashMap)transformCache;
                    LambdaFormEditor.Transform oldValue = (LambdaFormEditor.Transform)concurrentHashMap.putIfAbsent(withResult, withResult);
                    if (oldValue == null) {
                        return lambdaForm;
                    }
                    LambdaForm lambdaForm6 = (LambdaForm)oldValue.get();
                    if (lambdaForm6 != null) {
                        return lambdaForm6;
                    }
                    if (concurrentHashMap.replace(withResult, oldValue, withResult)) {
                        return lambdaForm;
                    }
                }
                else {
                    assert n == 0;
                    synchronized (this.lambdaForm) {
                        Object transformCache2 = this.lambdaForm.transformCache;
                        if (!(transformCache2 instanceof ConcurrentHashMap)) {
                            if (transformCache2 == null) {
                                this.lambdaForm.transformCache = withResult;
                                return lambdaForm;
                            }
                            LambdaFormEditor.Transform[] transformCache3;
                            if ((transformCache2 instanceof LambdaFormEditor.Transform)) {
                                LambdaFormEditor.Transform transform = (LambdaFormEditor.Transform)transformCache2;
                                if (transform.equals(withResult)) {
                                    LambdaForm lambdaForm2 = (LambdaForm)transform.get();
                                    if (lambdaForm2 == null) {
                                        this.lambdaForm.transformCache = withResult;
                                        return lambdaForm;
                                    }
                                    return lambdaForm2;
                                }
                                else {
                                    if (transform.get() == null) {
                                        this.lambdaForm.transformCache = withResult;
                                        return lambdaForm;
                                    }
                                    transformCache3 = new LambdaFormEditor.Transform[4];
                                    transformCache3[0] = transform;
                                    this.lambdaForm.transformCache = transformCache3;
                                }
                            }
                            else {
                                transformCache3 = (LambdaFormEditor.Transform[])(LambdaFormEditor.Transform[])transformCache2;
                            }
                            int length = transformCache3.length;
                            int n2 = -1;
                            int i = 0;
                            while (i < length) {
                                LambdaFormEditor.Transform transform2 = transformCache3[i];
                                if (transform2 == null) {
                                    break;
                                }
                                if (transform2.equals(withResult)) {
                                    LambdaForm lambdaForm4 = (LambdaForm)transform2.get();
                                    if (lambdaForm4 == null) {
                                        transformCache3[i] = withResult;
                                        return lambdaForm;
                                    }
                                    return lambdaForm4;
                                }
                                else {
                                    if (n2 < 0 && transform2.get() == null) {
                                        n2 = i;
                                    }
                                    ++i;
                                }
                            }
                            if (i >= length && n2 < 0) {
                                if (length < 16) {
                                    transformCache3 = (LambdaFormEditor.Transform[])Arrays.copyOf(transformCache3, Math.min(length * 2, 16));
                                    this.lambdaForm.transformCache = transformCache3;
                                }
                                else {
                                    ConcurrentHashMap transformCache4 = new ConcurrentHashMap(32);
                                    for (LambdaFormEditor.Transform transform3 : transformCache3) {
                                        transformCache4.put(transform3, transform3);
                                    }
                                    this.lambdaForm.transformCache = transformCache4;
                                    break Label_0518;
                                }
                            }
                            i = ((n2 >= 0) ? n2 : i);
                            transformCache3[i] = withResult;
                            return lambdaForm;
                        }
                    }
                }
            }
            ++n;
        }
    }
    
    private LambdaFormBuffer buffer() {
        return new LambdaFormBuffer(this.lambdaForm);
    }
    
    private BoundMethodHandle.SpeciesData oldSpeciesData() {
        return BoundMethodHandle.speciesData(this.lambdaForm);
    }
    
    private BoundMethodHandle.SpeciesData newSpeciesData(LambdaForm.BasicType basicType) {
        return this.oldSpeciesData().extendWith(basicType);
    }
    
    BoundMethodHandle bindArgumentL(BoundMethodHandle boundMethodHandle, int n, Object o) {
        assert boundMethodHandle.speciesData() == this.oldSpeciesData();
        return boundMethodHandle.copyWithExtendL(this.bindArgumentType(boundMethodHandle, n, LambdaForm.BasicType.L_TYPE), this.bindArgumentForm(1 + n), o);
    }
    
    BoundMethodHandle bindArgumentI(BoundMethodHandle boundMethodHandle, int n, int n2) {
        assert boundMethodHandle.speciesData() == this.oldSpeciesData();
        return boundMethodHandle.copyWithExtendI(this.bindArgumentType(boundMethodHandle, n, LambdaForm.BasicType.I_TYPE), this.bindArgumentForm(1 + n), n2);
    }
    
    BoundMethodHandle bindArgumentJ(BoundMethodHandle boundMethodHandle, int n, long n2) {
        assert boundMethodHandle.speciesData() == this.oldSpeciesData();
        return boundMethodHandle.copyWithExtendJ(this.bindArgumentType(boundMethodHandle, n, LambdaForm.BasicType.J_TYPE), this.bindArgumentForm(1 + n), n2);
    }
    
    BoundMethodHandle bindArgumentF(BoundMethodHandle boundMethodHandle, int n, float n2) {
        assert boundMethodHandle.speciesData() == this.oldSpeciesData();
        return boundMethodHandle.copyWithExtendF(this.bindArgumentType(boundMethodHandle, n, LambdaForm.BasicType.F_TYPE), this.bindArgumentForm(1 + n), n2);
    }
    
    BoundMethodHandle bindArgumentD(BoundMethodHandle boundMethodHandle, int n, double n2) {
        assert boundMethodHandle.speciesData() == this.oldSpeciesData();
        return boundMethodHandle.copyWithExtendD(this.bindArgumentType(boundMethodHandle, n, LambdaForm.BasicType.D_TYPE), this.bindArgumentForm(1 + n), n2);
    }
    
    private MethodType bindArgumentType(BoundMethodHandle boundMethodHandle, int n, LambdaForm.BasicType basicType) {
        assert boundMethodHandle.form.uncustomize() == this.lambdaForm;
        assert boundMethodHandle.form.names[1 + n].type == basicType;
        assert LambdaForm.BasicType.basicType(boundMethodHandle.type().parameterType(n)) == basicType;
        return boundMethodHandle.type().dropParameterTypes(n, n + 1);
    }
    
    LambdaForm bindArgumentForm(int n) {
        LambdaFormEditor.Transform of = LambdaFormEditor.Transform.of(LambdaFormEditor.Transform.Kind.BIND_ARG, n);
        LambdaForm inCache = this.getInCache(of);
        if (inCache != null) {
            assert inCache.parameterConstraint(0) == this.newSpeciesData(this.lambdaForm.parameterType(n));
            return inCache;
        }
        else {
            LambdaFormBuffer buffer = this.buffer();
            buffer.startEdit();
            BoundMethodHandle.SpeciesData oldSpeciesData = this.oldSpeciesData();
            BoundMethodHandle.SpeciesData speciesData = this.newSpeciesData(this.lambdaForm.parameterType(n));
            LambdaForm.Name parameter = this.lambdaForm.parameter(0);
            LambdaForm.NamedFunction getterFunction = speciesData.getterFunction(oldSpeciesData.fieldCount());
            if (n != 0) {
                buffer.replaceFunctions(oldSpeciesData.getterFunctions(), speciesData.getterFunctions(), new Object[] { parameter });
                LambdaForm.Name withConstraint2 = parameter.withConstraint(speciesData);
                buffer.renameParameter(0, withConstraint2);
                buffer.replaceParameterByNewExpression(n, new LambdaForm.Name(getterFunction, new Object[] { withConstraint2 }));
            }
            else {
                assert oldSpeciesData == BoundMethodHandle.SpeciesData.EMPTY;
                LambdaForm.Name withConstraint = new LambdaForm.Name(LambdaForm.BasicType.L_TYPE).withConstraint(speciesData);
                buffer.replaceParameterByNewExpression(0, new LambdaForm.Name(getterFunction, new Object[] { withConstraint }));
                buffer.insertParameter(0, withConstraint);
            }
            return this.putInCache(of, buffer.endEdit());
        }
    }
    
    LambdaForm addArgumentForm(int n, LambdaForm.BasicType basicType) {
        LambdaFormEditor.Transform of = LambdaFormEditor.Transform.of(LambdaFormEditor.Transform.Kind.ADD_ARG, n, basicType.ordinal());
        LambdaForm inCache = this.getInCache(of);
        if (inCache != null) {
            assert inCache.arity == this.lambdaForm.arity + 1;
            assert inCache.parameterType(n) == basicType;
            return inCache;
        }
        else {
            LambdaFormBuffer buffer = this.buffer();
            buffer.startEdit();
            buffer.insertParameter(n, new LambdaForm.Name(basicType));
            return this.putInCache(of, buffer.endEdit());
        }
    }
    
    LambdaForm dupArgumentForm(int n, int n2) {
        LambdaFormEditor.Transform of = LambdaFormEditor.Transform.of(LambdaFormEditor.Transform.Kind.DUP_ARG, n, n2);
        LambdaForm inCache = this.getInCache(of);
        if (inCache != null) {
            assert inCache.arity == this.lambdaForm.arity - 1;
            return inCache;
        }
        else {
            LambdaFormBuffer buffer = this.buffer();
            buffer.startEdit();
            assert this.lambdaForm.parameter(n).constraint == null;
            assert this.lambdaForm.parameter(n2).constraint == null;
            buffer.replaceParameterByCopy(n2, n);
            return this.putInCache(of, buffer.endEdit());
        }
    }
    
    LambdaForm spreadArgumentsForm(int p0, Class<?> p1, int p2) {
        // Bytecode:
        // 
        //     4: astore          4
        //     6: aload_2        
        //     7: astore          5
        //     9: aload           4
        //    11: invokevirtual   java/lang/Class.isPrimitive:()Z
        //    14: ifne            21
        //    17: ldc             [Ljava/lang/Object;.class
        //    19: astore          5
        //    21: aload           4
        //    23: invokestatic    java/lang/invoke/LambdaForm$BasicType.basicType:(Ljava/lang/Class;)Ljava/lang/invoke/LambdaForm$BasicType;
        //    26: astore          6
        //    28: aload           6
        //    30: invokevirtual   java/lang/invoke/LambdaForm$BasicType.ordinal:()I
        //    33: istore          7
        //    35: aload           6
        //    37: invokevirtual   java/lang/invoke/LambdaForm$BasicType.basicTypeClass:()Ljava/lang/Class;
        //    40: aload           4
        //    42: if_acmpeq       67
        //    45: aload           4
        //    47: invokevirtual   java/lang/Class.isPrimitive:()Z
        //    50: ifeq            67
        //    53: getstatic       java/lang/invoke/LambdaForm$BasicType.TYPE_LIMIT:I
        //    56: aload           4
        //    58: invokestatic    sun/invoke/util/Wrapper.forPrimitiveType:(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;
        //    61: invokevirtual   sun/invoke/util/Wrapper.ordinal:()I
        //    64: iadd           
        //    65: istore          7
        //    67: getstatic       java/lang/invoke/LambdaFormEditor$Transform$Kind.SPREAD_ARGS:Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;
        //    70: iload_1        
        //    71: iload           7
        //    73: iload_3        
        //    74: invokestatic    java/lang/invoke/LambdaFormEditor$Transform.of:(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;III)Ljava/lang/invoke/LambdaFormEditor$Transform;
        //    77: astore          8
        //    79: aload_0        
        //    80: aload           8
        //    82: invokespecial   java/lang/invoke/LambdaFormEditor.getInCache:(Ljava/lang/invoke/LambdaFormEditor$Transform;)Ljava/lang/invoke/LambdaForm;
        //    85: astore          9
        //    87: aload           9
        //    89: ifnull          128
        //    92: getstatic       java/lang/invoke/LambdaFormEditor.$assertionsDisabled:Z
        //    95: ifne            125
        //    98: aload           9
        //   100: getfield        java/lang/invoke/LambdaForm.arity:I
        //   103: aload_0        
        //   104: getfield        java/lang/invoke/LambdaFormEditor.lambdaForm:Ljava/lang/invoke/LambdaForm;
        //   107: getfield        java/lang/invoke/LambdaForm.arity:I
        //   110: iload_3        
        //   111: isub           
        //   112: iconst_1       
        //   113: iadd           
        //   114: if_icmpeq       125
        //   117: new             Ljava/lang/AssertionError;
        //   120: dup            
        //   121: invokespecial   java/lang/AssertionError.<init>:()V
        //   124: athrow         
        //   125: aload           9
        //   127: areturn        
        //   128: aload_0        
        //   129: invokespecial   java/lang/invoke/LambdaFormEditor.buffer:()Ljava/lang/invoke/LambdaFormBuffer;
        //   132: astore          10
        //   134: aload           10
        //   136: invokevirtual   java/lang/invoke/LambdaFormBuffer.startEdit:()V
        //   139: getstatic       java/lang/invoke/LambdaFormEditor.$assertionsDisabled:Z
        //   142: ifne            160
        //   145: iload_1        
        //   146: sipush          255
        //   149: if_icmple       160
        //   152: new             Ljava/lang/AssertionError;
        //   155: dup            
        //   156: invokespecial   java/lang/AssertionError.<init>:()V
        //   159: athrow         
        //   160: getstatic       java/lang/invoke/LambdaFormEditor.$assertionsDisabled:Z
        //   163: ifne            187
        //   166: iload_1        
        //   167: iload_3        
        //   168: iadd           
        //   169: aload_0        
        //   170: getfield        java/lang/invoke/LambdaFormEditor.lambdaForm:Ljava/lang/invoke/LambdaForm;
        //   173: getfield        java/lang/invoke/LambdaForm.arity:I
        //   176: if_icmple       187
        //   179: new             Ljava/lang/AssertionError;
        //   182: dup            
        //   183: invokespecial   java/lang/AssertionError.<init>:()V
        //   186: athrow         
        //   187: getstatic       java/lang/invoke/LambdaFormEditor.$assertionsDisabled:Z
        //   190: ifne            205
        //   193: iload_1        
        //   194: ifgt            205
        //   197: new             Ljava/lang/AssertionError;
        //   200: dup            
        //   201: invokespecial   java/lang/AssertionError.<init>:()V
        //   204: athrow         
        //   205: new             Ljava/lang/invoke/LambdaForm$Name;
        //   208: dup            
        //   209: getstatic       java/lang/invoke/LambdaForm$BasicType.L_TYPE:Ljava/lang/invoke/LambdaForm$BasicType;
        //   212: invokespecial   java/lang/invoke/LambdaForm$Name.<init>:(Ljava/lang/invoke/LambdaForm$BasicType;)V
        //   215: astore          11
        //   217: new             Ljava/lang/invoke/LambdaForm$Name;
        //   220: dup            
        //   221: getstatic       java/lang/invoke/MethodHandleImpl$Lazy.NF_checkSpreadArgument:Ljava/lang/invoke/LambdaForm$NamedFunction;
        //   224: iconst_2       
        //   225: anewarray       Ljava/lang/Object;
        //   228: dup            
        //   229: iconst_0       
        //   230: aload           11
        //   232: aastore        
        //   233: dup            
        //   234: iconst_1       
        //   235: iload_3        
        //   236: invokestatic    java/lang/Integer.valueOf:(I)Ljava/lang/Integer;
        //   239: aastore        
        //   240: invokespecial   java/lang/invoke/LambdaForm$Name.<init>:(Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/Object;)V
        //   243: astore          12
        //   245: aload_0        
        //   246: getfield        java/lang/invoke/LambdaFormEditor.lambdaForm:Ljava/lang/invoke/LambdaForm;
        //   249: invokevirtual   java/lang/invoke/LambdaForm.arity:()I
        //   252: istore          13
        //   254: aload           10
        //   256: iload           13
        //   258: iinc            13, 1
        //   261: aload           12
        //   263: invokevirtual   java/lang/invoke/LambdaFormBuffer.insertExpression:(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;
        //   266: pop            
        //   267: aload           5
        //   269: invokestatic    java/lang/invoke/MethodHandles.arrayElementGetter:(Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;
        //   272: astore          14
        //   274: iconst_0       
        //   275: istore          15
        //   277: iload           15
        //   279: iload_3        
        //   280: if_icmpge       345
        //   283: new             Ljava/lang/invoke/LambdaForm$Name;
        //   286: dup            
        //   287: aload           14
        //   289: iconst_2       
        //   290: anewarray       Ljava/lang/Object;
        //   293: dup            
        //   294: iconst_0       
        //   295: aload           11
        //   297: aastore        
        //   298: dup            
        //   299: iconst_1       
        //   300: iload           15
        //   302: invokestatic    java/lang/Integer.valueOf:(I)Ljava/lang/Integer;
        //   305: aastore        
        //   306: invokespecial   java/lang/invoke/LambdaForm$Name.<init>:(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)V
        //   309: astore          16
        //   311: aload           10
        //   313: iload           13
        //   315: iload           15
        //   317: iadd           
        //   318: aload           16
        //   320: invokevirtual   java/lang/invoke/LambdaFormBuffer.insertExpression:(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;
        //   323: pop            
        //   324: aload           10
        //   326: iload_1        
        //   327: iload           15
        //   329: iadd           
        //   330: iload           13
        //   332: iload           15
        //   334: iadd           
        //   335: invokevirtual   java/lang/invoke/LambdaFormBuffer.replaceParameterByCopy:(II)Ljava/lang/invoke/LambdaFormBuffer;
        //   338: pop            
        //   339: iinc            15, 1
        //   342: goto            277
        //   345: aload           10
        //   347: iload_1        
        //   348: aload           11
        //   350: invokevirtual   java/lang/invoke/LambdaFormBuffer.insertParameter:(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;
        //   353: pop            
        //   354: aload           10
        //   356: invokevirtual   java/lang/invoke/LambdaFormBuffer.endEdit:()Ljava/lang/invoke/LambdaForm;
        //   359: astore          9
        //   361: aload_0        
        //   362: aload           8
        //   364: aload           9
        //   366: invokespecial   java/lang/invoke/LambdaFormEditor.putInCache:(Ljava/lang/invoke/LambdaFormEditor$Transform;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/LambdaForm;
        //   369: areturn        
        //    Signature:
        //  (ILjava/lang/Class<*>;I)Ljava/lang/invoke/LambdaForm;
        //    StackMapTable: 00 09 FD 00 15 07 00 BD 07 00 BD FD 00 2D 07 00 C7 01 FD 00 39 07 00 CB 07 00 C6 02 FC 00 1F 07 00 CA 1A 11 FF 00 47 00 10 07 00 02 01 07 00 BD 01 07 00 BD 07 00 BD 07 00 C7 01 07 00 CB 07 00 C6 07 00 CA 07 00 C8 07 00 C8 01 07 00 CD 01 00 00 FA 00 43
        // 
        throw new IllegalStateException("An error occurred while decompiling this method.");
    }
    
    LambdaForm collectArgumentsForm(int n, MethodType methodType) {
        int parameterCount = methodType.parameterCount();
        int n2 = methodType.returnType() == Void.TYPE;
        if (parameterCount == 1 && n2 == 0) {
            return this.filterArgumentForm(n, LambdaForm.BasicType.basicType(methodType.parameterType(0)));
        }
        LambdaForm.BasicType[] basicTypes = LambdaForm.BasicType.basicTypes(methodType.parameterList());
        LambdaFormEditor.Transform.Kind kind = n2 ? LambdaFormEditor.Transform.Kind.COLLECT_ARGS_TO_VOID : LambdaFormEditor.Transform.Kind.COLLECT_ARGS;
        if (n2 && parameterCount == 0) {
            n = 1;
        }
        LambdaFormEditor.Transform of = LambdaFormEditor.Transform.of(kind, n, parameterCount, LambdaForm.BasicType.basicTypesOrd(basicTypes));
        LambdaForm inCache = this.getInCache(of);
        if (inCache != null) {
            assert inCache.arity == this.lambdaForm.arity - (n2 ? 0 : 1) + parameterCount;
            return inCache;
        }
        else {
            return this.putInCache(of, this.makeArgumentCombinationForm(n, methodType, false, (boolean)n2));
        }
    }
    
    LambdaForm collectArgumentArrayForm(int n, MethodHandle methodHandle) {
        MethodType type = methodHandle.type();
        int parameterCount = type.parameterCount();
        assert methodHandle.intrinsicName() == MethodHandleImpl.Intrinsic.NEW_ARRAY;
        Class<Object> componentType = type.returnType().getComponentType();
        LambdaForm.BasicType basicType = LambdaForm.BasicType.basicType(componentType);
        int ordinal = basicType.ordinal();
        if (basicType.basicTypeClass() != componentType) {
            if (!componentType.isPrimitive()) {
                return null;
            }
            ordinal = LambdaForm.BasicType.TYPE_LIMIT + Wrapper.forPrimitiveType(componentType).ordinal();
        }
        assert !(!type.parameterList().equals(Collections.nCopies(parameterCount, componentType)));
        LambdaFormEditor.Transform of = LambdaFormEditor.Transform.of(LambdaFormEditor.Transform.Kind.COLLECT_ARGS_TO_ARRAY, n, parameterCount, ordinal);
        LambdaForm inCache = this.getInCache(of);
        if (inCache != null) {
            assert inCache.arity == this.lambdaForm.arity - 1 + parameterCount;
            return inCache;
        }
        else {
            LambdaFormBuffer buffer = this.buffer();
            buffer.startEdit();
            assert n + 1 <= this.lambdaForm.arity;
            assert n > 0;
            LambdaForm.Name[] array = new LambdaForm.Name[parameterCount];
            for (int i = 0; i < parameterCount; ++i) {
                array[i] = new LambdaForm.Name(n + i, basicType);
            }
            LambdaForm.Name o = new LambdaForm.Name(methodHandle, (Object[])array);
            int arity = this.lambdaForm.arity();
            buffer.insertExpression(arity, o);
            int n2 = n + 1;
            for (int length = array.length, j = 0; j < length; ++j) {
                buffer.insertParameter(n2++, array[j]);
            }
            assert buffer.lastIndexOf(o) == arity + array.length;
            buffer.replaceParameterByCopy(n, arity + array.length);
            return this.putInCache(of, buffer.endEdit());
        }
    }
    
    LambdaForm filterArgumentForm(int n, LambdaForm.BasicType basicType) {
        LambdaFormEditor.Transform of = LambdaFormEditor.Transform.of(LambdaFormEditor.Transform.Kind.FILTER_ARG, n, basicType.ordinal());
        LambdaForm inCache = this.getInCache(of);
        if (inCache != null) {
            assert inCache.arity == this.lambdaForm.arity;
            assert inCache.parameterType(n) == basicType;
            return inCache;
        }
        else {
            return this.putInCache(of, this.makeArgumentCombinationForm(n, MethodType.methodType(this.lambdaForm.parameterType(n).basicTypeClass(), basicType.basicTypeClass()), false, false));
        }
    }
    
    private LambdaForm makeArgumentCombinationForm(int n, MethodType methodType, boolean b, boolean b2) {
        LambdaFormBuffer buffer = this.buffer();
        buffer.startEdit();
        int parameterCount = methodType.parameterCount();
        int n2 = b2 ? 0 : 1;
        assert n <= 255;
        assert n + n2 + (b ? parameterCount : 0) <= this.lambdaForm.arity;
        assert n > 0;
        assert methodType == methodType.basicType();
        assert !(!b2);
        BoundMethodHandle.SpeciesData oldSpeciesData = this.oldSpeciesData();
        BoundMethodHandle.SpeciesData speciesData = this.newSpeciesData(LambdaForm.BasicType.L_TYPE);
        LambdaForm.Name parameter = this.lambdaForm.parameter(0);
        buffer.replaceFunctions(oldSpeciesData.getterFunctions(), speciesData.getterFunctions(), new Object[] { parameter });
        LambdaForm.Name withConstraint = parameter.withConstraint(speciesData);
        buffer.renameParameter(0, withConstraint);
        LambdaForm.Name name = new LambdaForm.Name(speciesData.getterFunction(oldSpeciesData.fieldCount()), new Object[] { withConstraint });
        Object[] array = new Object[1 + parameterCount];
        array[0] = name;
        LambdaForm.Name[] array2;
        if (b) {
            array2 = new LambdaForm.Name[0];
            System.arraycopy(this.lambdaForm.names, n + n2, array, 1, parameterCount);
        }
        else {
            array2 = new LambdaForm.Name[parameterCount];
            LambdaForm.BasicType[] basicTypes = LambdaForm.BasicType.basicTypes(methodType.parameterList());
            for (int i = 0; i < basicTypes.length; ++i) {
                array2[i] = new LambdaForm.Name(n + i, basicTypes[i]);
            }
            System.arraycopy(array2, 0, array, 1, parameterCount);
        }
        LambdaForm.Name o = new LambdaForm.Name(methodType, array);
        int arity = this.lambdaForm.arity();
        buffer.insertExpression(arity + 0, name);
        buffer.insertExpression(arity + 1, o);
        int n3 = n + n2;
        for (int length = array2.length, j = 0; j < length; ++j) {
            buffer.insertParameter(n3++, array2[j]);
        }
        assert buffer.lastIndexOf(o) == arity + 1 + array2.length;
        if (!b2) {
            buffer.replaceParameterByCopy(n, arity + 1 + array2.length);
        }
        return buffer.endEdit();
    }
    
    LambdaForm filterReturnForm(LambdaForm.BasicType basicType, boolean b) {
        LambdaFormEditor.Transform of = LambdaFormEditor.Transform.of(b ? LambdaFormEditor.Transform.Kind.FILTER_RETURN_TO_ZERO : LambdaFormEditor.Transform.Kind.FILTER_RETURN, basicType.ordinal());
        LambdaForm inCache = this.getInCache(of);
        if (inCache != null) {
            assert inCache.arity == this.lambdaForm.arity;
            assert inCache.returnType() == basicType;
            return inCache;
        }
        else {
            LambdaFormBuffer buffer = this.buffer();
            buffer.startEdit();
            int length = this.lambdaForm.names.length;
            LambdaForm.Name result;
            if (b) {
                if (basicType == LambdaForm.BasicType.V_TYPE) {
                    result = null;
                }
                else {
                    result = new LambdaForm.Name(LambdaForm.constantZero(basicType), new Object[0]);
                }
            }
            else {
                BoundMethodHandle.SpeciesData oldSpeciesData = this.oldSpeciesData();
                BoundMethodHandle.SpeciesData speciesData = this.newSpeciesData(LambdaForm.BasicType.L_TYPE);
                LambdaForm.Name parameter = this.lambdaForm.parameter(0);
                buffer.replaceFunctions(oldSpeciesData.getterFunctions(), speciesData.getterFunctions(), new Object[] { parameter });
                LambdaForm.Name withConstraint = parameter.withConstraint(speciesData);
                buffer.renameParameter(0, withConstraint);
                LambdaForm.Name name = new LambdaForm.Name(speciesData.getterFunction(oldSpeciesData.fieldCount()), new Object[] { withConstraint });
                buffer.insertExpression(length++, name);
                LambdaForm.BasicType returnType = this.lambdaForm.returnType();
                if (returnType == LambdaForm.BasicType.V_TYPE) {
                    result = new LambdaForm.Name(MethodType.methodType(basicType.basicTypeClass()), new Object[] { name });
                }
                else {
                    result = new LambdaForm.Name(MethodType.methodType(basicType.basicTypeClass(), returnType.basicTypeClass()), new Object[] { name, this.lambdaForm.names[this.lambdaForm.result] });
                }
            }
            if (result != null) {
                buffer.insertExpression(length++, result);
            }
            buffer.setResult(result);
            return this.putInCache(of, buffer.endEdit());
        }
    }
    
    LambdaForm foldArgumentsForm(int n, boolean b, MethodType methodType) {
        int parameterCount = methodType.parameterCount();
        LambdaFormEditor.Transform.Kind kind = b ? LambdaFormEditor.Transform.Kind.FOLD_ARGS_TO_VOID : LambdaFormEditor.Transform.Kind.FOLD_ARGS;
        LambdaFormEditor.Transform of = LambdaFormEditor.Transform.of(kind, n, parameterCount);
        LambdaForm inCache = this.getInCache(of);
        if (inCache != null) {
            assert inCache.arity == this.lambdaForm.arity - ((kind == LambdaFormEditor.Transform.Kind.FOLD_ARGS) ? 1 : 0);
            return inCache;
        }
        else {
            return this.putInCache(of, this.makeArgumentCombinationForm(n, methodType, true, b));
        }
    }
    
    LambdaForm permuteArgumentsForm(int n, int[] array) {
        assert n == 1;
        int length = this.lambdaForm.names.length;
        int length2 = array.length;
        int max = 0;
        boolean n2 = true;
        int n3;
        for (int i = 0; i < array.length; max = Math.max(max, n3 + 1), ++i) {
            n3 = array[i];
            if (n3 != i) {
                n2 = false;
            }
        }
        assert n + array.length == this.lambdaForm.arity;
        if (!(n2 == 0)) {
            return this.lambdaForm;
        }
        LambdaFormEditor.Transform of = LambdaFormEditor.Transform.of(LambdaFormEditor.Transform.Kind.PERMUTE_ARGS, array);
        LambdaForm inCache = this.getInCache(of);
        if (inCache != null) {
            assert inCache.arity == n + max : inCache;
            return inCache;
        }
        else {
            LambdaForm.BasicType[] array2 = new LambdaForm.BasicType[max];
            for (int j = 0; j < length2; ++j) {
                array2[array[j]] = this.lambdaForm.names[n + j].type;
            }
            assert n + length2 == this.lambdaForm.arity;
            assert !(!permutedTypesMatch(array, array2, this.lambdaForm.names, n));
            int k;
            for (k = 0; k < length2 && array[k] == k; ++k) {}
            LambdaForm.Name[] array3 = new LambdaForm.Name[length - length2 + max];
            System.arraycopy(this.lambdaForm.names, 0, array3, 0, n + k);
            int n4 = length - this.lambdaForm.arity;
            System.arraycopy(this.lambdaForm.names, n + length2, array3, n + max, n4);
            int l = array3.length - n4;
            int result = this.lambdaForm.result;
            if (result >= n) {
                if (result < n + length2) {
                    result = array[result - n] + n;
                }
                else {
                    result = result - length2 + max;
                }
            }
            for (; k < length2; ++k) {
                LambdaForm.Name name = this.lambdaForm.names[n + k];
                int n5 = array[k];
                LambdaForm.Name name2 = array3[n + n5];
                if (name2 == null) {
                    name2 = (array3[n + n5] = new LambdaForm.Name(array2[n5]));
                }
                else
                    assert name2.type == array2[n5];
                for (; l < array3.length; ++l) {
                    array3[l] = array3[l].replaceName(name, name2);
                }
            }
            for (int n6 = n + k; n6 < l; ++n6) {
                if (array3[n6] == null) {
                    array3[n6] = LambdaForm.argument(n6, array2[n6 - n]);
                }
            }
            for (int arity = this.lambdaForm.arity; arity < this.lambdaForm.names.length; ++arity) {
                int n7 = arity - this.lambdaForm.arity + l;
                LambdaForm.Name name3 = this.lambdaForm.names[arity];
                LambdaForm.Name name4 = array3[n7];
                if (name3 != name4) {
                    for (int n8 = n7 + 1; n8 < array3.length; ++n8) {
                        array3[n8] = array3[n8].replaceName(name3, name4);
                    }
                }
            }
            return this.putInCache(of, new LambdaForm(this.lambdaForm.debugName, l, array3, result));
        }
    }
    
    static boolean permutedTypesMatch(int[] array, LambdaForm.BasicType[] array2, LambdaForm.Name[] array3, int n) {
        int i = 0;
        while (i < array.length) {
            assert !(!array3[n + i].isParam());
            assert array3[n + i].type == array2[array[i]];
            ++i;
        }
        return true;
    }
    
    private static final class Transform extends SoftReference<LambdaForm>
    {
        final long packedBytes;
        final byte[] fullBytes;
        private static final boolean STRESS_TEST = false;
        private static final int PACKED_BYTE_SIZE = 4;
        private static final int PACKED_BYTE_MASK = 15;
        private static final int PACKED_BYTE_MAX_LENGTH = 16;
        private static final byte[] NO_BYTES;
        
        private static long packedBytes(byte[] array) {
            if (array.length > 16) {
                return 0L;
            }
            long n = 0L;
            int n2 = 0;
            for (int i = 0; i < array.length; ++i) {
                int n3 = array[i] & 0xFF;
                n2 |= n3;
                n |= (long)n3 << i * 4;
            }
            if (!inRange(n2)) {
                return 0L;
            }
            return n;
        }
        
        private static long packedBytes(int n, int n2) {
            assert !(!inRange(n | n2));
            return (long)(n << 0 | n2 << 4);
        }
        
        private static long packedBytes(int n, int n2, int n3) {
            assert !(!inRange(n | n2 | n3));
            return (long)(n << 0 | n2 << 4 | n3 << 8);
        }
        
        private static long packedBytes(int n, int n2, int n3, int n4) {
            assert !(!inRange(n | n2 | n3 | n4));
            return (long)(n << 0 | n2 << 4 | n3 << 8 | n4 << 12);
        }
        
        private static boolean inRange(int n) {
            assert (n & 0xFF) == n;
            return (n & 0xFFFFFFF0) == 0x0;
        }
        
        private static byte[] fullBytes(int... array) {
            byte[] array2 = new byte[array.length];
            int n = 0;
            for (int length = array.length, i = 0; i < length; ++i) {
                array2[n++] = bval(array[i]);
            }
            assert packedBytes(array2) == 0L;
            return array2;
        }
        
        private byte byteAt(int n) {
            long packedBytes = this.packedBytes;
            if (packedBytes == 0L) {
                if (n >= this.fullBytes.length) {
                    return 0;
                }
                return this.fullBytes[n];
            }
            else {
                assert this.fullBytes == null;
                if (n > 16) {
                    return 0;
                }
                return (byte)(int)(packedBytes >>> n * 4 & 0xFL);
            }
        }
        
        LambdaFormEditor.Transform.Kind kind() {
            return LambdaFormEditor.Transform.Kind.values()[this.byteAt(0)];
        }
        
        private Transform(long packedBytes, byte[] fullBytes, LambdaForm referent) {
            super(referent);
            this.packedBytes = packedBytes;
            this.fullBytes = fullBytes;
        }
        
        private Transform(long n) {
            this(n, null, null);
            assert n != 0L;
            return;
        }
        
        private Transform(byte[] array) {
            this(0L, array, null);
        }
        
        private static byte bval(int n) {
            assert (n & 0xFF) == n;
            return (byte)n;
        }
        
        private static byte bval(LambdaFormEditor.Transform.Kind kind) {
            return bval(kind.ordinal());
        }
        
        static LambdaFormEditor.Transform of(LambdaFormEditor.Transform.Kind kind, int n) {
            int bval = bval(kind);
            if (inRange(bval | n)) {
                return new LambdaFormEditor.Transform(packedBytes(bval, n));
            }
            return new LambdaFormEditor.Transform(fullBytes(new int[] { bval, n }));
        }
        
        static LambdaFormEditor.Transform of(LambdaFormEditor.Transform.Kind kind, int n, int n2) {
            int n3 = (byte)kind.ordinal();
            if (inRange(n3 | n | n2)) {
                return new LambdaFormEditor.Transform(packedBytes(n3, n, n2));
            }
            return new LambdaFormEditor.Transform(fullBytes(new int[] { n3, n, n2 }));
        }
        
        static LambdaFormEditor.Transform of(LambdaFormEditor.Transform.Kind kind, int n, int n2, int n3) {
            int n4 = (byte)kind.ordinal();
            if (inRange(n4 | n | n2 | n3)) {
                return new LambdaFormEditor.Transform(packedBytes(n4, n, n2, n3));
            }
            return new LambdaFormEditor.Transform(fullBytes(new int[] { n4, n, n2, n3 }));
        }
        
        static LambdaFormEditor.Transform of(LambdaFormEditor.Transform.Kind kind, int... array) {
            return ofBothArrays(kind, array, LambdaFormEditor.Transform.NO_BYTES);
        }
        
        static LambdaFormEditor.Transform of(LambdaFormEditor.Transform.Kind kind, int n, byte[] array) {
            return ofBothArrays(kind, new int[] { n }, array);
        }
        
        static LambdaFormEditor.Transform of(LambdaFormEditor.Transform.Kind kind, int n, int n2, byte[] array) {
            return ofBothArrays(kind, new int[] { n, n2 }, array);
        }
        
        private static LambdaFormEditor.Transform ofBothArrays(LambdaFormEditor.Transform.Kind kind, int[] array, byte[] array2) {
            byte[] array3 = new byte[1 + array.length + array2.length];
            int n = 0;
            array3[n++] = bval(kind);
            for (int length = array.length, i = 0; i < length; ++i) {
                array3[n++] = bval(array[i]);
            }
            for (int length2 = array2.length, j = 0; j < length2; ++j) {
                array3[n++] = array2[j];
            }
            long packedBytes = packedBytes(array3);
            if (packedBytes != 0L) {
                return new Transform(packedBytes);
            }
            return new Transform(array3);
        }
        
        LambdaFormEditor.Transform withResult(LambdaForm lambdaForm) {
            return new LambdaFormEditor.Transform(this.packedBytes, this.fullBytes, lambdaForm);
        }
        
        @Override
        public boolean equals(Object o) {
            return o instanceof LambdaFormEditor.Transform && this.equals((LambdaFormEditor.Transform)o);
        }
        
        public boolean equals(LambdaFormEditor.Transform transform) {
            return this.packedBytes == transform.packedBytes && Arrays.equals(this.fullBytes, transform.fullBytes);
        }
        
        @Override
        public int hashCode() {
            if (this.packedBytes != 0L) {
                assert this.fullBytes == null;
                return Long.hashCode(this.packedBytes);
            }
            else {
                return Arrays.hashCode(this.fullBytes);
            }
        }
        
        @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();
            long packedBytes = this.packedBytes;
            if (packedBytes != 0L) {
                sb.append("(");
                while (packedBytes != 0L) {
                    sb.append(packedBytes & 0xFL);
                    packedBytes >>>= 4;
                    if (packedBytes == 0L) {
                        continue;
                    }
                    sb.append(",");
                }
                sb.append(")");
            }
            if (this.fullBytes != null) {
                sb.append("unpacked");
                sb.append(Arrays.toString(this.fullBytes));
            }
            LambdaForm obj = (LambdaForm)this.get();
            if (obj != null) {
                sb.append(" result=");
                sb.append(obj);
            }
            return sb.toString();
        }
        
        static {
            NO_BYTES = new byte[0];
        }
        
        private enum Kind
        {
            NO_KIND, 
            BIND_ARG, 
            ADD_ARG, 
            DUP_ARG, 
            SPREAD_ARGS, 
            FILTER_ARG, 
            FILTER_RETURN, 
            FILTER_RETURN_TO_ZERO, 
            COLLECT_ARGS, 
            COLLECT_ARGS_TO_VOID, 
            COLLECT_ARGS_TO_ARRAY, 
            FOLD_ARGS, 
            FOLD_ARGS_TO_VOID, 
            PERMUTE_ARGS;
        }
    }
}
